//
// Licensed under Apache License v2.0.  See the file "LICENSE" for more
// information.
//

#ifndef _CUPS_FILTERS_PDFTOPDF_QPDF_XOBJECT_H_
#define _CUPS_FILTERS_PDFTOPDF_QPDF_XOBJECT_H_

#include <qpdf/QPDFObjectHandle.hh>

QPDFObjectHandle _cfPDFToPDFMakeXObject(QPDF *pdf, QPDFObjectHandle page);

#endif // !_CUPS_FILTERS_PDFTOPDF_QPDF_XOBJECT_H_
